#ifndef SEMANTICS_H_
#define SEMANTICS_H_

#include <map>

#include "Pcl6BaseVisitor.h"
#include "antlr4-runtime.h"

#include "intermediate/symtab/SymtabStack.h"
#include "intermediate/symtab/SymtabEntry.h"
#include "intermediate/symtab/Predefined.h"
#include "intermediate/type/Typespec.h"
#include "intermediate/util/BackendMode.h"
#include "SemanticErrorHandler.h"

namespace frontend {

using namespace std;
using namespace intermediate::symtab;
using namespace intermediate::type;

class Semantics : public Pcl6BaseVisitor
{
private:
    BackendMode mode;
    SymtabStack *symtabStack;
    SymtabEntry *programId;
    SemanticErrorHandler error;

    map<string, Typespec *> *typeTable;

    /**
     * Return the number of values in a datatype.
     * @param type the datatype.
     * @return the number of values.
     */
    int typeCount(Typespec *type);

    /**
     * Determine whether or not an expression is a variable only.
     * @param exprCtx the ExpressionContext.
     * @return true if it's an expression only, else false.
     */
    bool expressionIsVariable(Pcl6Parser::ExpressionContext *exprCtx);

    /**
     * Perform semantic operations on procedure and function call arguments.
     * @param listCtx the ArgumentListContext.
     * @param parameters the vector of parameters to fill.
     */
//    void checkCallArguments(Pcl6Parser::ArgumentListContext *listCtx,
//                            vector<SymtabEntry *> *parms);

    /**
     * Determine the datatype of a variable that can have modifiers.
     * @param varCtx the VariableContext.
     * @param varType the variable's datatype without the modifiers.
     * @return the datatype with any modifiers.
     */
    Typespec *variableDatatype(Pcl6Parser::VariableContext *varCtx,
                               Typespec *varType);

    /**
     * Create a new record type.
     * @param recordTypeSpecCtx the RecordTypespecContext.
     * @param recordTypeName the name of the record type.
     * @return the symbol table entry of the record type identifier.
     */
    SymtabEntry *createRecordType(
                        Pcl6Parser::RecordTypespecContext *recordTypeSpecCtx,
                        string recordTypeName);

    /**
     * Create the fully qualified type pathname of a record type.
     * @param recordType the record type.
     * @return the pathname.
     */
    string createRecordTypePath(Typespec *recordType);

    /**
     * Create the symbol table for a record type.
     * @param ctx the RecordFieldsContext,
     * @param ownerId the symbol table entry of the owner's identifier.
     * @return the symbol table.
     */
    Symtab *createRecordSymtab(
                Pcl6Parser::RecordFieldsContext *ctx, SymtabEntry *ownerId);

public:
    Semantics(BackendMode mode) : mode(mode), programId(nullptr)
    {
        // Create and initialize the symbol table stack.
        symtabStack = new SymtabStack();
        Predefined::initialize(symtabStack);

        typeTable = new map<string, Typespec *>();
        (*typeTable)["integer"] = Predefined::integerType;
        (*typeTable)["real"]    = Predefined::realType;
        (*typeTable)["boolean"] = Predefined::booleanType;
        (*typeTable)["char"]    = Predefined::charType;
        (*typeTable)["string"]  = Predefined::stringType;
    }

    /**
     * Get the symbol table entry of the program identifier.
     * @return the entry.
     */
    SymtabEntry *getProgramId() { return programId; }

    /**
     * Get the count of semantic errors.
     * @return the count.
     */
    int getErrorCount() const { return error.getCount(); }

    /**
     * Return the default value for a given datatype.
     * @param type the datatype.
     */
    static Object defaultValue(Typespec *type);

    Object visitProgram(Pcl6Parser::ProgramContext *ctx) override;
    Object visitProgramHeader(Pcl6Parser::ProgramHeaderContext *ctx) override;
    Object visitConstantDefinition(Pcl6Parser::ConstantDefinitionContext *ctx) override;
    Object visitConstant(Pcl6Parser::ConstantContext *ctx) override;
    Object visitTypeDefinition(Pcl6Parser::TypeDefinitionContext *ctx) override;
    Object visitRecordTypespec(Pcl6Parser::RecordTypespecContext *ctx) override;
    Object visitSimpleTypespec(Pcl6Parser::SimpleTypespecContext *ctx) override;
    Object visitTypeIdentifierTypespec(Pcl6Parser::TypeIdentifierTypespecContext *ctx) override;
    Object visitTypeIdentifier(Pcl6Parser::TypeIdentifierContext *ctx) override;
    Object visitEnumerationTypespec(Pcl6Parser::EnumerationTypespecContext *ctx) override;
    Object visitSubrangeTypespec(Pcl6Parser::SubrangeTypespecContext *ctx) override;
    Object visitArrayTypespec(Pcl6Parser::ArrayTypespecContext *ctx) override;
    Object visitVariableDeclarations(Pcl6Parser::VariableDeclarationsContext *ctx) override;
    Object visitAssignmentStatement(Pcl6Parser::AssignmentStatementContext *ctx) override;
    Object visitLhs(Pcl6Parser::LhsContext *ctx) override;
    Object visitIfStatement(Pcl6Parser::IfStatementContext *ctx) override;
    Object visitCaseStatement(Pcl6Parser::CaseStatementContext *ctx) override;
    Object visitRepeatStatement(Pcl6Parser::RepeatStatementContext *ctx) override;
    Object visitWhileStatement(Pcl6Parser::WhileStatementContext *ctx) override;
    Object visitForStatement(Pcl6Parser::ForStatementContext *ctx) override;
    Object visitExpression(Pcl6Parser::ExpressionContext *ctx) override;
    Object visitSimpleExpression(Pcl6Parser::SimpleExpressionContext *ctx) override;
    Object visitTerm(Pcl6Parser::TermContext *ctx) override;
    Object visitVariableFactor(Pcl6Parser::VariableFactorContext *ctx) override;
    Object visitVariable(Pcl6Parser::VariableContext *ctx) override;
    Object visitVariableIdentifier(Pcl6Parser::VariableIdentifierContext *ctx) override;
    Object visitNumberFactor(Pcl6Parser::NumberFactorContext *ctx) override;
    Object visitCharacterFactor(Pcl6Parser::CharacterFactorContext *ctx) override;
    Object visitStringFactor(Pcl6Parser::StringFactorContext *ctx) override;
    Object visitNotFactor(Pcl6Parser::NotFactorContext *ctx) override;
    Object visitParenthesizedFactor(Pcl6Parser::ParenthesizedFactorContext *ctx) override;
};

} // namespace frontend

#endif /* SEMANTICS_H_ */
